"""
MIT License

(C) 2024-2025 ramalama developers

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
"""

import subprocess
import glob
import platform

class GPUDetector:
    def __init__(self):
        self.best_gpu = None
        self.best_vram = 0
        self.best_env = None

    def _update_best_gpu(self, memory_mib, gpu_name, env_var):
        """Updates the best available GPU based on highest VRAM."""
        if memory_mib > 1024 and memory_mib > self.best_vram:
            self.best_vram = memory_mib
            self.best_gpu = gpu_name
            self.best_env = env_var

    def _read_gpu_memory(self, paths, gpu_prefix, env_var):
        """Reads VRAM from sysfs for AMD and Intel GPUs and updates best GPU."""
        for i, memory_path in enumerate(sorted(glob.glob(paths))):
            try:
                with open(memory_path, 'r') as file:
                    memory_bytes = int(file.read().strip())
                    memory_mib = memory_bytes // (1024 * 1024)  # Convert bytes to MiB
                    self._update_best_gpu(memory_mib, f"{gpu_prefix} {i}", env_var)
            except Exception as ex:
                raise RuntimeError(f"Error reading {gpu_prefix} GPU memory from {memory_path}: {ex}")

    def get_nvidia_gpu(self):
        """Detects Nvidia GPUs using nvidia-smi (Linux only)."""
        if platform.system() != "Linux":
            return  # Skip on macOS and other platforms

        try:
            result = subprocess.run(
                ["nvidia-smi", "--query-gpu=index,memory.total", "--format=csv,noheader,nounits"],
                capture_output=True, text=True, check=True
            )
            output = result.stdout.strip()

            for line in output.split('\n'):
                try:
                    index, memory_mib = line.split(',')
                    memory_mib = int(memory_mib.strip())
                    self._update_best_gpu(memory_mib, index.strip(), "CUDA_VISIBLE_DEVICES")
                except ValueError:
                    raise RuntimeError(f"Error parsing Nvidia GPU info: {line}")

        except FileNotFoundError:
            raise RuntimeError("`nvidia-smi` not found. No NVIDIA GPU detected or drivers missing.")
        except subprocess.CalledProcessError as e:
            error_msg = e.stderr.strip() if e.stderr else "Unknown error (check if NVIDIA drivers are loaded)."
            raise RuntimeError(f"Unable to detect NVIDIA GPU(s). Error: {error_msg}")

    def get_amd_gpu(self):
        """Detects AMD GPUs using sysfs on Linux or system_profiler on macOS."""
        if platform.system() == "Linux":
            self._read_gpu_memory('/sys/bus/pci/devices/*/mem_info_vram_total', "AMD GPU", "HIP_VISIBLE_DEVICES")
        elif platform.system() == "Darwin":  # macOS
            self.get_macos_gpu()  # macOS detection covers AMD GPUs

    def get_intel_gpu(self):
        """Detects Intel GPUs using sysfs and lspci (Linux) or system_profiler (macOS)."""
        if platform.system() == "Linux":
            self._read_gpu_memory('/sys/class/drm/card*/device/mem_info_vram_total', "Intel GPU", "ONEAPI_DEVICE_SELECTOR")

            # Alternative detection method using lspci
            try:
                result = subprocess.run(["lspci", "-nn"], capture_output=True, text=True, check=True)
                output = result.stdout.strip()
                if "Intel Corporation" in output:
                    if self.best_gpu is None:  # If no VRAM info was found
                        self.best_gpu = "Intel GPU (unknown memory)"
                        self.best_env = "ONEAPI_DEVICE_SELECTOR"
            except FileNotFoundError:
                raise RuntimeError("`lspci` not found. Unable to detect Intel GPU.")
            except subprocess.CalledProcessError as e:
                error_msg = e.stderr.strip() if e.stderr else "Unknown error (check if `pciutils` is installed)."
                raise RuntimeError(f"Unable to detect Intel GPU(s). Error: {error_msg}")
        elif platform.system() == "Darwin":  # macOS
            self.get_macos_gpu()  # macOS detection covers Intel GPUs

    def get_macos_gpu(self):
        """Detects GPUs on macOS using system_profiler."""
        try:
            result = subprocess.run(["system_profiler", "SPDisplaysDataType"], capture_output=True, text=True, check=True)
            output = result.stdout.strip()

            gpu_name = None
            vram_mib = 0  # Default to unknown

            for line in output.split('\n'):
                line = line.strip()
                if line.startswith("Chipset Model:"):
                    gpu_name = line.split(":")[1].strip()
                elif line.startswith("VRAM") or line.startswith("VRAM (Dynamic)"):
                    try:
                        vram_mib = int(line.split(":")[1].strip().replace("MB", ""))
                        self._update_best_gpu(vram_mib, gpu_name, "METAL_DEVICE_INDEX")
                    except ValueError:
                        continue  # Skip invalid VRAM values

            # Handle Apple Silicon (M1, M2, etc.)
            if gpu_name and "Apple" in gpu_name:
                self._update_best_gpu(8192, gpu_name, "METAL_DEVICE_INDEX")  # Assume 8GB VRAM for M1/M2 Max

        except FileNotFoundError:
            raise RuntimeError("`system_profiler` not found. Unable to detect macOS GPU.")
        except subprocess.CalledProcessError as e:
            error_msg = e.stderr.strip() if e.stderr else "Unknown error."
            raise RuntimeError(f"Unable to detect macOS GPU(s). Error: {error_msg}")

    def detect_best_gpu(self, gpu_template):
        """Compares Nvidia, AMD, and Intel GPUs and appends the best GPU with the highest VRAM to gpu_template.
        If one type of GPU fails, it continues to the next type.
        """
        system = platform.system()

        if system == "Linux":
            try:
                self.get_nvidia_gpu()
            except RuntimeError as e:
                print(f"Warning: {e}")  # Log error but continue

            try:
                self.get_amd_gpu()
            except RuntimeError as e:
                print(f"Warning: {e}")  # Log error but continue

            try:
                self.get_intel_gpu()
            except RuntimeError as e:
                print(f"Warning: {e}")  # Log error but continue

        elif system == "Darwin":  # macOS
            try:
                self.get_macos_gpu()
            except RuntimeError as e:
                print(f"Warning: {e}")  # Log error but continue

        else:
            raise RuntimeError(f"GPU detection is not supported on {system}.")

        if self.best_gpu is not None:
            gpu_template.append({
                "index": self.best_gpu,
                "vram": self.best_vram,
                "env": self.best_env
            })
            return True  # GPU detected and added
        else:
            raise RuntimeError("No compatible GPUs found.")
